#include "stdio.h"
#include <ctime>

// Device code
__global__ void VecAdd(const float* A, const float* B, float* C, int N)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (threadIdx.x < 10)
    {
        C[i] = A[i] + B[i];
    }
    else
    {
        C[i] = A[i] - B[i];
    }
}

// Host code
void Init(float* data, int end)
{
    for (int i = 0; i < end; ++i)
    {
      data[i] = rand()/ (float)RAND_MAX;
      printf("%f ", data[i]);
    }
}

int main(int argc, char** argv)
{
    srand(time(0));
    int N = 1000;
    size_t size = N * sizeof(float);

    // Allocate input vectors h_A and h_B in host memory
    float* h_A;
    float* h_B;
    h_A = (float*)malloc(size);
    h_B = (float*)malloc(size);
    
    // Allocate vectors in device memory
    float* d_A;
    float* d_B;
    float* d_C;
    cudaMalloc((void**)&d_A, size);
    cudaMalloc((void**)&d_B, size);
    cudaMalloc((void**)&d_C, size);

    // Initialize input vectors
    Init(h_A, N);
    Init(h_B, N);

    // Copy vectors from host memory to device memory
    cudaMemcpy(d_A, h_A, size, cudaMemcpyHostToDevice); 
    cudaMemcpy(d_B, h_B, size, cudaMemcpyHostToDevice);

    // Invoke kernel
    int threadsPerBlock = 256;
    int blocksPerGrid = (N + threadsPerBlock - 1) / threadsPerBlock;
    VecAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, N);

  return 0;
}


